"""Asynchronous Python client for NET2GRID devices."""
from __future__ import annotations

import asyncio
import json
import socket
from dataclasses import dataclass
from importlib import metadata
from typing import Any, Self, cast

from aiohttp import ClientError, ClientSession
from aiohttp.hdrs import METH_GET
from yarl import URL

from .exceptions import GridNetConnectionError
from .models import Device, SmartBridge


@dataclass
class GridNet:
    """Main class for handling connections with the devices."""

    host: str
    request_timeout: float = 10.0
    session: ClientSession | None = None

    _close_session: bool = False

    async def _request(
        self,
        uri: str,
        *,
        method: str = METH_GET,
        data: dict[str, Any] | None = None,
    ) -> Any:
        """Handle a request to the device.

        Args:
        ----
            uri: Request URI, without '/', for example, 'status'
            method: HTTP Method to use.
            data: Dictionary of data to send to the API.

        Returns:
        -------
            A Python dictionary (text) with the response from
            a device.

        Raises:
        ------
            GridNetConnectionError: An error occurred while
                communicating with the device.
        """
        version = metadata.version(__package__)
        url = URL.build(scheme="http", host=self.host, path="/").join(URL(uri))

        headers = {
            "User-Agent": f"PythonGridNet/{version}",
            "Accept": "application/json, text/plain, */*",
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.request(
                    method,
                    url,
                    json=data,
                    headers=headers,
                )
                response.raise_for_status()
        except asyncio.TimeoutError as exception:
            msg = f"Timeout occurred while connecting to {self.host}"
            raise GridNetConnectionError(
                msg,
            ) from exception
        except (
            ClientError,
            socket.gaierror,
        ) as exception:
            msg = f"Error occurred while communicating with {self.host}"
            raise GridNetConnectionError(
                msg,
            ) from exception

        return cast(dict[str, Any], json.loads(await response.text()))

    async def device(self) -> Device:
        """Get the latest values from a the device.

        Returns
        -------
            A Device data object from the API.
        """
        data = await self._request("info")
        return Device.from_dict(data)

    async def smartbridge(self) -> SmartBridge:
        """Get the latest values from a the device.

        Returns
        -------
            A SmartBridge data object from the API.
        """
        data = await self._request("meter/now")
        return SmartBridge.from_dict(data)

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The GridNet object.
        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.
        """
        await self.close()



# pylint: disable=W0621
"""Asynchronous Python client for a NET2GRID device."""

import asyncio

from gridnet import Device, GridNet


async def main() -> None:
    """Test device output."""
    async with GridNet(
        host="127.0.0.1",
    ) as gridnet:
        device: Device = await gridnet.device()
        print(device)
        print(f"ID: {device.n2g_id}")
        print(f"Model: {device.model}")
        print(f"Batch: {device.batch}")
        print(f"Firmware version: {device.firmware}")
        print(f"Hardware version: {device.hardware}")
        print(f"Manufacturer: {device.manufacturer}")


if __name__ == "__main__":
    asyncio.run(main())



"""Models for GridNet."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class SmartBridge:
    """Object representing an SmartBridge response from the device."""

    power_flow: int
    energy_consumption_total: float
    energy_production_total: float

    @staticmethod
    def from_dict(data: dict[str, Any]) -> SmartBridge:
        """Return SmartBridge object from the API response.

        Args:
        ----
            data: Response data from the API.

        Returns:
        -------
            A SmartBridge object.
        """
        data = data["elec"]

        def convert(value: float) -> float:
            """Convert the unit of measurement.

            Args:
            ----
                value: input value.

            Returns:
            -------
                Value in kWh rounded with 1 decimal.
            """
            value = value / 1000
            return float(round(value, 1))

        return SmartBridge(
            power_flow=data["power"]["now"].get("value"),
            energy_consumption_total=convert(data["import"]["now"].get("value")),
            energy_production_total=convert(data["export"]["now"].get("value")),
        )


@dataclass
class Device:
    """Object representing an Device response from the device."""

    n2g_id: str
    model: str
    batch: str
    firmware: str
    hardware: int
    manufacturer: str

    @staticmethod
    def from_dict(data: dict[str, Any]) -> Device:
        """Return Device object from the API response.

        Args:
        ----
            data: Response data from the API.

        Returns:
        -------
            A Device object.
        """
        return Device(
            n2g_id=data["id"],
            model=data["model"],
            batch=data["batch"],
            firmware=data["fw"],
            hardware=data["hw"],
            manufacturer=data["mf"],
        )



# pylint: disable=W0621
"""Asynchronous Python client for a NET2GRID device."""

import asyncio

from gridnet import GridNet, SmartBridge


async def main() -> None:
    """Test smartbridge output."""
    async with GridNet(
        host="127.0.0.1",
    ) as gridnet:
        smartbridge: SmartBridge = await gridnet.smartbridge()
        print(smartbridge)
        print()
        print(f"Power flow: {smartbridge.power_flow}")
        print(f"Energy consumption: {smartbridge.energy_consumption_total}")
        print(f"Energy production: {smartbridge.energy_production_total}")


if __name__ == "__main__":
    asyncio.run(main())



"""Asynchronous Python client for a NED2GRID device."""

from .exceptions import GridNetConnectionError, GridNetError
from .gridnet import GridNet
from .models import Device, SmartBridge

__all__ = [
    "Device",
    "GridNet",
    "GridNetError",
    "GridNetConnectionError",
    "SmartBridge",
]



"""Basic tests for the API."""
# pylint: disable=protected-access
import asyncio
from unittest.mock import patch

import pytest
from aiohttp import ClientError, ClientResponse, ClientSession
from aresponses import Response, ResponsesMockServer

from gridnet import GridNet
from gridnet.exceptions import GridNetConnectionError, GridNetError

from . import load_fixtures


async def test_json_request(aresponses: ResponsesMockServer) -> None:
    """Test JSON response is handled correctly."""
    aresponses.add(
        "127.0.0.1",
        "/test",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text='{"status": "ok"}',
        ),
    )
    async with ClientSession() as session:
        client = GridNet("127.0.0.1", session=session)
        await client._request("test")
        await client.close()


async def test_internal_session(aresponses: ResponsesMockServer) -> None:
    """Test JSON response is handled correctly."""
    aresponses.add(
        "127.0.0.1",
        "/test",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text='{"status": "ok"}',
        ),
    )
    async with GridNet("127.0.0.1") as client:
        await client._request("test")


async def test_timeout(aresponses: ResponsesMockServer) -> None:
    """Test request timeout from the API."""

    # Faking a timeout by sleeping
    async def response_handler(_: ClientResponse) -> Response:
        await asyncio.sleep(0.2)
        return aresponses.Response(
            body="Goodmorning!",
            text=load_fixtures("smartbridge.json"),
        )

    aresponses.add("127.0.0.1", "/meter/now", "GET", response_handler)

    async with ClientSession() as session:
        client = GridNet(host="127.0.0.1", session=session, request_timeout=0.1)
        with pytest.raises(GridNetConnectionError):
            assert await client.smartbridge()


async def test_client_error() -> None:
    """Test request client error from the API."""
    async with ClientSession() as session:
        client = GridNet(host="127.0.0.1", session=session)
        with patch.object(
            session,
            "request",
            side_effect=ClientError,
        ), pytest.raises(GridNetConnectionError):
            assert await client._request("test")


@pytest.mark.parametrize("status", [401, 403])
async def test_http_error401(aresponses: ResponsesMockServer, status: int) -> None:
    """Test HTTP 401 response handling."""
    aresponses.add(
        "127.0.0.1",
        "/meter/now",
        "GET",
        aresponses.Response(text="Give me energy!", status=status),
    )

    async with ClientSession() as session:
        client = GridNet(host="127.0.0.1", session=session)
        with pytest.raises(GridNetError):
            assert await client._request("test")


async def test_http_error400(aresponses: ResponsesMockServer) -> None:
    """Test HTTP 404 response handling."""
    aresponses.add(
        "127.0.0.1",
        "/meter/now",
        "GET",
        aresponses.Response(text="Give me energy!", status=404),
    )

    async with ClientSession() as session:
        client = GridNet(host="127.0.0.1", session=session)
        with pytest.raises(GridNetError):
            assert await client._request("test")


async def test_http_error500(aresponses: ResponsesMockServer) -> None:
    """Test HTTP 500 response handling."""
    aresponses.add(
        "127.0.0.1",
        "/meter/now",
        "GET",
        aresponses.Response(
            body=b'{"status":"nok"}',
            status=500,
        ),
    )

    async with ClientSession() as session:
        client = GridNet(host="127.0.0.1", session=session)
        with pytest.raises(GridNetError):
            assert await client._request("test")


async def test_no_success(aresponses: ResponsesMockServer) -> None:
    """Test a message without a success message throws."""
    aresponses.add(
        "127.0.0.1",
        "/meter/now",
        "GET",
        aresponses.Response(
            status=200,
            text='{"message": "no success"}',
        ),
    )

    async with ClientSession() as session:
        client = GridNet(host="127.0.0.1", session=session)
        with pytest.raises(GridNetError):
            assert await client._request("test")



"""Test the models."""
from aiohttp import ClientSession
from aresponses import ResponsesMockServer

from gridnet import Device, GridNet, SmartBridge

from . import load_fixtures


async def test_device(aresponses: ResponsesMockServer) -> None:
    """Test request from the device - Device object."""
    aresponses.add(
        "127.0.0.1",
        "/info",
        "GET",
        aresponses.Response(
            text=load_fixtures("device.json"),
            status=200,
        ),
    )

    async with ClientSession() as session:
        client = GridNet(host="127.0.0.1", session=session)
        device: Device = await client.device()
        assert device
        assert device.n2g_id == "84df:0c11:9999:3795"
        assert device.model == "SBWF3102"
        assert device.batch == "SBP-HMX-210318"
        assert device.firmware == "1.6.16"
        assert device.hardware == 1
        assert device.manufacturer == "NET2GRID"


async def test_smartbridge(aresponses: ResponsesMockServer) -> None:
    """Test request from the device - SmartBridge object."""
    aresponses.add(
        "127.0.0.1",
        "/meter/now",
        "GET",
        aresponses.Response(
            text=load_fixtures("smartbridge.json"),
            status=200,
        ),
    )

    async with ClientSession() as session:
        client = GridNet(host="127.0.0.1", session=session)
        smartbridge: SmartBridge = await client.smartbridge()
        assert smartbridge
        assert smartbridge.power_flow == 338
        assert smartbridge.energy_consumption_total == 17762.1
        assert smartbridge.energy_production_total == 21214.6
